#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Machine Learning éthique
# 
# Ce notebook est inspiré de l'article [FairTest: Discovering Unwarranted Associations in Data-Driven Applications](https://arxiv.org/abs/1510.02377) et propose d'étudier une façon de vérifier qu'un modèle ou une métrique n'est pas biaisé par rapport à certains critères.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Principe
# 
# Un problème de machine learning dans sa définition la plus simple consiste à prédire $Y$ en fonction de $I$. On souhaite construire une fonction $f(I)$ qui s'approche le plus possible de $Y$. On appelle $I$ l'ensemble des données utilisateurs ou *user input* et $f(I)$ la sortie du modèle ou *output to users (O)*. On veut découvrir si la sortie du modèle est biaisée d'une façon non-éthique par rapport à des attributs dits protégés ou *protected attributes (S)* tels que l'origine éthnique ou le genre. On tient compte également du contexte, code postal, ...) ou *context attributes (X)* et de variables explicative ou *explanotory attributes (E)*. On a donc :
# 
# * **I** : les variables en entrée du modèles,
# * **O** : la prédiction du modèle,
# * **S** : les attributs protégés, toute corrélation entre la sortie et ces attributs est indésirable,
# * **X** : le contexte (principalement le lieu géographique)
# * **E** : les variables explicatives (contraintes qualitification), la sortie est a priori corrélée avec ces attributs, de façon quasi prévisible mais cela reste éthique.
# 
# Un exemple. Supposons qu'on cherche à recommander un objet à la vente sur un site internet à partir des données de navigation de cette personne (I). Comme cette personne est identifiée, on peut savoir où elle habite (X), son origine éthnique (S) et son niveau d'étude (E). Le site lui recommande un livre de mathématiques (O). Est-ce que cette recommandation dépend principalement de l'origine éthnique (S) ? Et si les deux informations semblent corrélées, ne serait-ce pas plutôt son niveau d'étude (E) qui explique la recommandation ?
# 
# Concrètement, la sortie *O* ne dépend que de *I* mais on souhaite savoir si *I* ne serait pas corrélées à certains facteurs qui ne sont pas éthique. C'est une corrélation inattendue mais observable. Le contexte *X* permet de partitionner et de vérifier si le modèle est éthique sur l'ensemble des sous-groupes de population. Le processus de recherche des corrélations indésirées s'effectue en cinq temps.
# 
# * On choisit un attribut protégé $S_i$ et on réalise une partition de la population pour cet attribut.
# * On vérifie pour chaque sous-population (ou chaque partie) que la sortie *O* et l'attribut $S_i$ ne sont pas corrélés. C'est l'étape de **détection**.
# * Pour chaque biais détecté, on vérifie qu'il ne peut être expliqué par une des variables explicatrice auquel ce biais est acceptable. C'est l'étape de **debugging**.

# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# ## Installation de FairTest
# 
# Le module ne peut pas être installé tel quel car il a été implémenté pour Python 2. Voici un lien qui devrait vous permettre de l'installer pour Python 3 si vous ne l'avez pas déjà fait.

# In[3]:


# !pip install https://github.com/sdpython/fairtest/releases/download/0.1/fairtest-0.1-py3-none-any.whl


# ## Données
# 
# On récupère les données [adult](https://archive.ics.uci.edu/ml/datasets/Adult) pour lequel il faut prédire le fait qu'une personne ait un revenu supérieur à 50.000 dollars par an.

# In[4]:


import pyensae.datasource as ds
data = ds.download_data("adult.data",
                        url="https://archive.ics.uci.edu/ml/machine-learning-databases/adult/")
names = ("age,workclass,fnlwgt,education,education-num,marital-status,occupation,relationship,"
         "race,sexe,capital-gain,capital-loss,hours-per-week,native-country,income").split(",")


# In[5]:


import pandas
df = pandas.read_csv(data, names=names)
df.head()


# ## Exemple avec aequitas
# 
# [aequitas](https://github.com/dssg/aequitas) propose un ensemble de métrique pour mesurer les biais éthiques.

# In[6]:


from sklearn.model_selection import train_test_split
df2 = df.copy()
df2_train, df2_test = train_test_split(df2)


# In[7]:


import numpy
label = "income"
X_train = df2_train.drop(label, axis=1)
y_train = df2_train[label] == ' >50K'
y_train = pandas.Series(numpy.array([1.0 if y else 0.0 for y in y_train]))
X_test = df2_test.drop(label, axis=1)
y_test = df2_test[label] == ' >50K'
y_test = pandas.Series(numpy.array([1.0 if y else 0.0 for y in y_test]))


# In[8]:


X_train = X_train.drop(['fnlwgt'], axis=1).copy()
X_test = X_test.drop(['fnlwgt'], axis=1).copy()


# In[9]:


cat_col = list(_ for _ in X_train.select_dtypes("object").columns)
cat_col


# In[10]:


from sklearn.tree import DecisionTreeClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import LabelEncoder, OneHotEncoder



pipe = make_pipeline(
            OneHotEncoder(handle_unknown="ignore"),
            DecisionTreeClassifier(min_samples_leaf=10))
pipe.fit(X_train, y_train)


# In[11]:


data = X_test.copy().reset_index(drop=True)
data["score"] = pipe.predict_proba(X_test)[:, 1]
data["label_value"] = y_test
data.head(n=4)


# In[12]:


data_small = data[["sexe", "education", "score", "label_value"]].copy()
data_small.head(n=5)


# In[13]:


from seaborn import catplot, violinplot
g = violinplot(x="sexe", y="score", hue="label_value", data=data_small)


# In[14]:


from aequitas.group import Group
g = Group()
xtab, _ = g.get_crosstabs(data_small)


# In[15]:


xtab


# In[16]:


from aequitas.plotting import Plot

aqp = Plot()
fpr_plot = aqp.plot_group_metric(xtab, 'fpr')


# In[17]:


set(data_small['education'])


# In[18]:


from aequitas.bias import Bias

b = Bias()
bdf = b.get_disparity_predefined_groups(xtab, 
                    original_df=data_small, 
                    ref_groups_dict={'sexe': ' Male', 'education':' Masters'},
                    alpha=0.05, 
                    check_significance=False)


# In[19]:


fpr_disparity = aqp.plot_disparity(bdf, group_metric='fpr_disparity', 
                                   attribute_name='education')


# ## Exemple avec fairtest
# 
# Ne fonctionne pas encore. Il faut passer à Python trois.

# In[20]:


try:
    from fairtest import DataSource
    cont = True
except ImportError:
    cont = False
# On prend les 1000 premières lignes.
# Le processus est assez long.
if cont:
    dsdat = DataSource(df[:1000], budget=1, conf=0.95)


# In[21]:


if cont:
    from fairtest import Testing, train, test, report

    SENS = ['sex', 'race']     # Protected features
    TARGET = 'income'             # Output
    EXPL = ''                     # Explanatory feature

    inv = Testing(dsdat, SENS, TARGET, EXPL)
    train([inv])


# In[22]:


if cont:
    test([inv])


# In[23]:


if cont:
    try:
        report([inv], "adult")
    except Exception as e:
        print("ERROR")
        print(e)
        # à corriger encore


# ## Exercice 1 : Construire un arbre de décision avec son propre critère
# 
# On pourra s'inspirer de l'article [Pure Python Decision Trees](http://kldavenport.com/pure-python-decision-trees/),
# ou [How To Implement The Decision Tree Algorithm From Scratch In Python](https://machinelearningmastery.com/implement-decision-tree-algorithm-scratch-python/). Cet arbre de décision doit produire un résultat similaire à ceci : [Contexte-education-num-in-(9.5, 11.5)-age-in(46.5, inf)](https://github.com/sdpython/fairtest/blob/master/results/report_adult.txt#L35). Il s'agit d'implémenter l'algorithme qui suit avec la métrique [Mutual Information (MI)](https://fr.wikipedia.org/wiki/Information_mutuelle).

# In[24]:


from pyquickhelper.helpgen import NbImage
NbImage("fairtesttree.png")


# In[25]:





# ## Exercice 2 : appliquer l'algorithme sur le jeu de données adulte

# In[26]:





# ## Exercice 3 : apprendre sans interactions

# In[27]:


df2 = df.copy()
df2["un"] = 1
df2["age10"] = (df["age"] // 10) * 10
gr = df2[["age10", "sexe", "income", "un"]].groupby(["age10", "sexe", "income"], as_index=False).sum()
gr.head()


# In[28]:


g = gr.pivot_table("un", "age10", ["income", "sexe"])
g


# In[29]:


g.columns


# In[30]:


g["rF"] = g[(" <=50K", " Female")] / g[(" >50K", " Female")]
g["rM"] = g[(" <=50K", " Male")] / g[(" >50K", " Male")]
g


# Les deux variables paraissent corrélées.

# On apprend trois modèles :
#     
# * On prédit 'income' avec l'âge et le genre, modèle M0
# * On prédit 'income' avec l'âge uniquement, modèle M1
# * On prédit 'income' avec le genre uniquement, modèle M2
# * On prédit 'income' avec les modèles M1 et M2.
# 
# Le modèle prédit-il aussi bien, moins bien ? Qu'en déduisez-vous ?

# In[31]: